SystemOrganization addCategory: #'Xpf-Multipoint-Core'!

!MenuMorph methodsFor: '*xpf-multipoint-core-override' stamp: 'bs 5/20/2009 23:43'!
popUpAt: aPoint forHand: hand in: aWorld allowKeyboard: aBoolean 
	"Present this menu at the given point under control of the given  
	hand."
	| evt |
	hand menus select: [:menu | menu stayUp not] 
				  thenCollect: [:menu | menu delete. hand menus remove: menu].
	
	hand menus add: self. "a weakset is used"

	self items isEmpty
		ifTrue: [^ self].

	MenuIcons decorateMenu: self.

	(self submorphs
		select: [:m | m isKindOf: UpdatingMenuItemMorph])
		do: [:m | m updateContents].
	"precompute width"
	self
		positionAt: aPoint
		relativeTo: (selectedItem
				ifNil: [self items first])
		inWorld: aWorld.
	aWorld addMorphFront: self.
	"Acquire focus for valid pop up behavior"
	hand newMouseFocus: self.
	aBoolean
		ifTrue: [hand newKeyboardFocus: self].
	evt := hand lastEvent.
	(evt isKeyboard
			or: [evt isMouse
					and: [evt anyButtonPressed not]])
		ifTrue: ["Select first item if button not down"
			self moveSelectionDown: 1 event: evt].
	self updateColor.
	self changed! !

Object subclass: #MultiPointer
	instanceVariableNames: 'pointerID'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

MultiPointer class
	instanceVariableNames: 'eventSemaphore deviceSemaphore'!

!MultiPointer commentStamp: 'mig 5/19/2009 19:49' prior: 0!
This class provides abstractions to the MultiPointer plugin. Instances are bound to a specific pointer in the operating system and provide messages for retrieving information about them.!

MultiPointer class
	instanceVariableNames: 'eventSemaphore deviceSemaphore'!

!MultiPointer class methodsFor: 'instance creation' stamp: 'mig 5/19/2009 10:51'!
allPointers

	^ (0 to: self getPointerCount - 1) collect: [:id | self newForID: id]! !

!MultiPointer class methodsFor: 'accessing' stamp: 'mig 5/6/2009 22:53'!
deviceSemaphore

	^ deviceSemaphore ifNil: [self deviceSemaphore: Semaphore new. deviceSemaphore]! !

!MultiPointer class methodsFor: 'accessing' stamp: 'mig 5/6/2009 22:07'!
deviceSemaphore: anObject

	deviceSemaphore := anObject.! !

!MultiPointer class methodsFor: 'accessing' stamp: 'mig 5/6/2009 22:53'!
eventSemaphore

	^ eventSemaphore ifNil: [self eventSemaphore: Semaphore new. eventSemaphore]! !

!MultiPointer class methodsFor: 'accessing' stamp: 'mig 5/6/2009 22:07'!
eventSemaphore: anObject

	eventSemaphore := anObject.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'mig 5/6/2009 22:10'!
getPointerCount

	<primitive: 'primitiveMultiPointerCount' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'mig 5/19/2009 10:44'!
getPointerCurrentState: aPointerID

	^ MultiPointerState new in: [:state | 
		self pvtGetPointerCurrentState: aPointerID with: state. state]
! !

!MultiPointer class methodsFor: 'primitives' stamp: 'mig 5/6/2009 22:30'!
getPointerLively: aPointerID

	^ (self pvtGetPointerLively: aPointerID) ~= 0! !

!MultiPointer class methodsFor: 'primitives' stamp: 'mig 4/22/2009 17:34'!
getPointerName: aPointerID

	| pointerName |
	pointerName := self pvtGetPointerName: aPointerID.
	^ pointerName truncateTo: (pointerName 
			indexOf: (Character value: 0) 
			startingAt: 0 
			ifAbsent: [pointerName size + 1]) - 1! !

!MultiPointer class methodsFor: 'instance creation' stamp: 'mig 5/19/2009 10:51'!
newForID: aPointerID

	^ self new pointerID: aPointerID! !

!MultiPointer class methodsFor: 'primitives' stamp: 'mig 5/6/2009 22:11'!
pvtGetPointerCurrentState: aPointerID with: aMultiPointerState

	<primitive: 'primitiveMultiPointerCurrentState' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'mig 5/6/2009 22:11'!
pvtGetPointerLively: aPointerID

	<primitive: 'primitiveMultiPointerLively' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'mig 5/6/2009 22:11'!
pvtGetPointerName: aPointerID

	<primitive: 'primitiveMultiPointerName' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'system startup' stamp: 'mig 5/19/2009 12:10'!
registerSemaphores

	self setPointerEventSemaphoreHandle: (Smalltalk registerExternalObject: self eventSemaphore).
	self setPointerDeviceSemaphoreHandle: (Smalltalk registerExternalObject: self deviceSemaphore).! !

!MultiPointer class methodsFor: 'accessing' stamp: 'mig 5/19/2009 14:58'!
resetAll

	self allPointers do: [:pointer | pointer isLively ifTrue: [pointer reset]].! !

!MultiPointer class methodsFor: 'primitives' stamp: 'mig 5/6/2009 22:11'!
setPointerDeviceSemaphoreHandle: aSemaIndex

	<primitive: 'primitiveMultiPointerDeviceSemaphoreHandle' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'mig 5/6/2009 22:11'!
setPointerEventSemaphoreHandle: aSemaIndex

	<primitive: 'primitiveMultiPointerEventSemaphoreHandle' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'system startup' stamp: 'mig 5/19/2009 12:10'!
unregisterSemaphores

	self setPointerEventSemaphoreHandle: 0.
	self setPointerDeviceSemaphoreHandle: 0.
	
	Smalltalk unregisterExternalObject: self eventSemaphore.
	Smalltalk unregisterExternalObject: self deviceSemaphore.
	
	self eventSemaphore: nil.
	self deviceSemaphore: nil.
! !

!MultiPointer class methodsFor: 'primitives' stamp: 'mig 5/6/2009 22:12'!
useHardwareMouse

	<primitive: 'primitiveMultiPointerUseHardwareMouse' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'mig 5/6/2009 22:12'!
useSoftwareMouse

	<primitive: 'primitiveMultiPointerUseSoftwareMouse' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer methodsFor: 'comparing' stamp: 'mig 5/19/2009 10:58'!
= aPointer

	^ self pointerID = aPointer pointerID! !

!MultiPointer methodsFor: 'testing' stamp: 'mig 5/19/2009 10:13'!
isLively

	^ self class getPointerLively: self pointerID! !

!MultiPointer methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:12'!
pointerID

	^ pointerID! !

!MultiPointer methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:19'!
pointerID: anInteger

	pointerID := anInteger.! !

!MultiPointer methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:48'!
reset

	self class getPointerCurrentState: self pointerID.! !

!MultiPointer methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:18'!
state

	^ self class getPointerCurrentState: self pointerID! !

Object subclass: #MultiPointerManager
	instanceVariableNames: 'handMorphs keyboardHand deviceUpdateProcess movementProcess colors'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

MultiPointerManager class
	instanceVariableNames: 'instance'!

!MultiPointerManager commentStamp: 'mig 5/19/2009 19:47' prior: 0!
A MultiPointerManager controls all MultiPointer actions. There should be only one MultiPointerManager instance in a running system. Instances are created and deleted upon startup and shutdown.

colors: is a dictionary having colors as keys and usage counts as values.
deviceUpdateProcess and movementProcess: hold the two forked processes for MultiPointer.
handMorphs: is a list of all multipointer hands.
keyboardHand: is the hand which sends keyboard events into the world.
!

MultiPointerManager class
	instanceVariableNames: 'instance'!

!MultiPointerManager class methodsFor: 'class initialization' stamp: 'mig 5/19/2009 11:44'!
initialize

	super initialize.
	Smalltalk addToStartUpList: self.
	Smalltalk addToShutDownList: self.
	! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:20'!
instance

	^ instance! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:20'!
instance: anObject

	instance := anObject.! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:37'!
multiPointerActive

	^ self instance notNil and: [self instance multiPointerActive]! !

!MultiPointerManager class methodsFor: 'testing' stamp: 'SR 5/6/2009 10:19'!
pluginAvailable

	^ SmalltalkImage current listBuiltinModules anySatisfy: [:each | each startsWith: 'MultiPointerPlugin']! !

!MultiPointerManager class methodsFor: 'system startup' stamp: 'mig 5/19/2009 11:55'!
shutDown: aboutToQuit 

	self instance ifNotNil: [
		self instance shutdown.
		self instance: nil].
	
	super shutDown: aboutToQuit.! !

!MultiPointerManager class methodsFor: 'event processing' stamp: 'mig 5/19/2009 11:57'!
startMoving

	self instance ifNotNil: [ self instance startMoving ]! !

!MultiPointerManager class methodsFor: 'system startup' stamp: 'mig 5/19/2009 11:54'!
startUp: aboutToQuit 

	self pluginAvailable ifTrue: [
		self instance: self new.
		self instance startup].
	
	super startUp: aboutToQuit.! !

!MultiPointerManager class methodsFor: 'event processing' stamp: 'mig 5/19/2009 11:57'!
stopMoving

	self instance ifNotNil: [ self instance stopMoving ]! !

!MultiPointerManager methodsFor: 'hands' stamp: 'test 5/20/2009 13:28'!
addAllHands

	self handMorphs do: [:hand | 	
		World addHand: hand.
		hand changed].
	self keyboardHand: self handMorphs first.
! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 16:19'!
colors

	^ colors! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 16:19'!
colors: anObject

	colors := anObject.! !

!MultiPointerManager methodsFor: 'hands' stamp: 'mig 5/19/2009 20:33'!
createHandMorphFor: aPointer
	
	| hand |
	(self handMorphs anySatisfy: [:each | each pointer = aPointer]) 
		ifTrue: [^ self].

	hand := MultiPointerHandMorph new 
		pointer: aPointer;
		handColor: self getColor.
	self handMorphs add: hand.
	self multiPointerActive ifTrue: [World addHand: hand].
! !

!MultiPointerManager methodsFor: 'hands' stamp: 'mig 5/19/2009 16:32'!
destroyHandMorphFor: aPointer

	(self handMorphs detect: [:each | each pointer = aPointer] ifNone: [nil])
		ifNotNilDo: [:morph |
			self handMorphs remove: morph.
			World removeHand: morph.
			self releaseColor: morph handColor.
			
			self keyboardHand = morph ifTrue: [
				self handMorphs
					ifEmpty: [self keyboardHand: World hands first]
					ifNotEmpty: [self keyboardHand: self handMorphs first]]].
! !

!MultiPointerManager methodsFor: 'processes' stamp: 'mig 5/19/2009 16:18'!
deviceUpdateBlock

	^ [[
			MultiPointer allPointers do: [:pointer |
				pointer isLively
					ifTrue: [ self createHandMorphFor: pointer ]
					ifFalse: [ self destroyHandMorphFor: pointer ]].
			MultiPointer deviceSemaphore wait.
		] repeat ]! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:23'!
deviceUpdateProcess

	^ deviceUpdateProcess! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:23'!
deviceUpdateProcess: anObject

	deviceUpdateProcess := anObject.! !

!MultiPointerManager methodsFor: 'hand colors' stamp: 'mig 5/19/2009 16:26'!
getColor

	| color |
	color := self colors keyAtValue: self colors values min.
	self colors at: color put: (self colors at: color) + 1.
	^ color! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:21'!
handMorphs

	^ handMorphs! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:21'!
handMorphs: anObject

	handMorphs := anObject.! !

!MultiPointerManager methodsFor: 'initialize-release' stamp: 'mig 5/19/2009 16:20'!
initialize

	super initialize.
	
	self handMorphs: OrderedCollection new.
	self initializeColors.! !

!MultiPointerManager methodsFor: 'hand colors' stamp: 'mig 5/19/2009 16:44'!
initializeColors

	self colors: (Dictionary newFrom: {
		Color red -> 0.
		Color orange -> 0.
		Color green -> 0.
		Color blue -> 0.
	}).
! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:22'!
keyboardHand

	^ keyboardHand! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 16:18'!
keyboardHand: anObject

	keyboardHand := anObject.
! !

!MultiPointerManager methodsFor: 'processes' stamp: 'mig 5/19/2009 11:23'!
movementBlock

	^ [[
			MultiPointer eventSemaphore wait.
			self handMorphs do: [:each | each readEvent]
		] repeat ]! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:23'!
movementProcess

	^ movementProcess! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:23'!
movementProcess: anObject

	movementProcess := anObject.! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 14:44'!
multiPointerActive

	^ self movementProcess notNil! !

!MultiPointerManager methodsFor: 'hand colors' stamp: 'mig 5/19/2009 16:28'!
releaseColor: aColor

	self colors at: aColor put: (self colors at: aColor) - 1! !

!MultiPointerManager methodsFor: 'hands' stamp: 'mig 5/19/2009 16:46'!
removeAllHands

	self keyboardHand: nil.
	self handMorphs copy do: [:each | World removeHand: each].
! !

!MultiPointerManager methodsFor: 'system startup' stamp: 'mig 5/19/2009 20:55'!
shutdown

	self stopMoving.
	self stopDeviceDiscovery.
	MultiPointer unregisterSemaphores.! !

!MultiPointerManager methodsFor: 'processes' stamp: 'mig 5/19/2009 15:14'!
startDeviceDiscovery

	self deviceUpdateProcess ifNotNil: [^self].
	
	self deviceUpdateProcess: 
		(self deviceUpdateBlock
			forkAt: Processor userBackgroundPriority
			named: 'MultiPointer Device Discovery').
	! !

!MultiPointerManager methodsFor: 'processes' stamp: 'mig 5/19/2009 15:13'!
startMoving

	self movementProcess ifNotNil: [^ self].
	
	self addAllHands.

	MultiPointer resetAll; useSoftwareMouse.
	
	self movementProcess:
		(self movementBlock
			forkAt: Processor userInterruptPriority
			named: 'MultiPointer Movement').
! !

!MultiPointerManager methodsFor: 'system startup' stamp: 'mig 5/19/2009 12:10'!
startup

	MultiPointer registerSemaphores.
	self startDeviceDiscovery.! !

!MultiPointerManager methodsFor: 'processes' stamp: 'mig 5/19/2009 11:28'!
stopDeviceDiscovery

	self deviceUpdateProcess ifNotNil: [
		self deviceUpdateProcess terminate.
		self deviceUpdateProcess: nil.
		
		MultiPointer allPointers do: [:each |
			self destroyHandMorphFor: each].
		
		self handMorphs: OrderedCollection new].
! !

!MultiPointerManager methodsFor: 'processes' stamp: 'mig 5/19/2009 15:29'!
stopMoving

	self movementProcess ifNotNil: [
		self movementProcess terminate.
		self movementProcess: nil].
	
	self removeAllHands.
	
	MultiPointer useHardwareMouse.
! !

Object subclass: #MultiPointerState
	instanceVariableNames: 'x y buttons'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

!MultiPointerState commentStamp: 'mig 5/19/2009 12:04' prior: 0!
MultiPointerState is a wrapper class for the status of a pointer reported by the MultiPointer plugin.
Do not change the instance variable definition because instance variables are set via index by MultiPointerPlugin.!

!MultiPointerState methodsFor: 'accessing' stamp: 'SR 4/24/2009 15:37'!
buttons

	^ buttons! !

!MultiPointerState methodsFor: 'accessing' stamp: 'SR 4/24/2009 15:37'!
x

	^ x! !

!MultiPointerState methodsFor: 'accessing' stamp: 'SR 4/24/2009 15:37'!
y

	^ y! !

Object subclass: #MultiPointerWidget
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

!MultiPointerWidget class methodsFor: 'user interface' stamp: 'SR 5/5/2009 23:18'!
createButton: actionSel label: labelSel
	
	| btn |
	btn := UITheme current
		newLabelIn: World
		label: labelSel. 
	btn 
		vResizing: #spaceFill; 
		hResizing: #spaceFill.
	btn on: #mouseUp send: actionSel to: self.
	^ btn
	! !

!MultiPointerWidget class methodsFor: 'user interface' stamp: 'mig 5/19/2009 20:26'!
grabKeyboardWithEvent: evt from: source

	MultiPointerManager multiPointerActive
		ifTrue: [MultiPointerManager instance keyboardHand: evt hand].! !

!MultiPointerWidget class methodsFor: 'user interface' stamp: 'mig 5/6/2009 22:25'!
open
	
	(AlignmentMorph newBounds: (50@50 corner: 700@80))
		color: Color yellow;
		addMorph: (self createButton: #grabKeyboardWithEvent:from: label: 'Grab Keyboard');
		openInWorld.! !

!HandMorph methodsFor: '*xpf-multipoint-core' stamp: 'bs 5/20/2009 23:36'!
menus
	^ self valueOfProperty: #xpfMultipointMenus ifAbsentPut: WeakSet new! !

!HandMorph methodsFor: '*xpf-multipoint-core-override' stamp: 'mig 5/19/2009 19:43'!
processEvents
	"Process user input events from the local input devices."
	| evtBuf hadAny shouldProcessEvents |

	shouldProcessEvents := [MultiPointerManager multiPointerActive not] ifError: [true].
	
	shouldProcessEvents ifTrue: [
		ActiveEvent ifNotNil: [
			"Meaning that we were invoked from within an event response. Make sure z-order is up to date"
			self mouseOverHandler processMouseOver: lastMouseEvent]].
	
	hadAny := false.
	
	[(evtBuf := Sensor nextEvent) notNil]
		whileTrue: [ | evt type |
			type := evtBuf first.
			evt := nil.
			
			shouldProcessEvents 
				ifTrue: [
					"normal behaviour"
					type = EventTypeMouse ifTrue: [evt := self generateMouseEvent: evtBuf].
					type = EventTypeKeyboard ifTrue: [evt := self generateKeyboardEvent: evtBuf].
					type = EventTypeDragDropFiles ifTrue: [evt := self generateDropFilesEvent: evtBuf].
				
					evt ifNotNil: [self handleEvent: evt. hadAny := true]]
				ifFalse: [
					"multipointer behaviour: process keyboard only and forward to another hand"
					type = EventTypeKeyboard ifTrue: [evt := self generateKeyboardEvent: evtBuf].
					
					"sometimes even generateKeyboardEvent returns mouse events (for mouse wheels)"
					(evt notNil and: [evt isKeyboard]) ifTrue: [
							MultiPointerManager instance keyboardHand enqueueEvent: evt]]].
	
	
	shouldProcessEvents ifTrue: [
		"note: if we come here we didn't have any mouse events"
		mouseClickState ifNotNil: [
			"No mouse events during this cycle. Make sure click states time out accordingly"
			mouseClickState handleEvent: lastMouseEvent asMouseMove from: self].
	
		hadAny ifFalse: ["No pending events. Make sure z-order is up to date"
			self mouseOverHandler processMouseOver: lastMouseEvent]].
! !

HandMorph subclass: #MultiPointerHandMorph
	instanceVariableNames: 'queue lastPosition buttons pointer normalCursor handColor movedTimestamp'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

!MultiPointerHandMorph class methodsFor: 'misc' stamp: 'test 5/20/2009 14:46'!
cursorColored: aColor

	^ self scaledCursor: 35 colored: aColor! !

!MultiPointerHandMorph class methodsFor: 'as yet unclassified' stamp: 'mig 5/27/2009 10:56'!
displayTimeout

	^ 5000! !

!MultiPointerHandMorph class methodsFor: 'icons' stamp: 'SR 5/25/2009 15:51'!
keyboardIcon

	^ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self keyboardIconContents readStream)! !

!MultiPointerHandMorph class methodsFor: 'icons' stamp: 'SR 5/25/2009 17:29'!
keyboardIconContents

	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJ
bWFnZVJlYWR5ccllPAAAAdxJREFUeNrEU0tLG1EU/iadgggVMUJomhQxPiuU+CjURTZWEJFuuhLd
uelfCP6AbsWlP6BQcVtaqC/sxkWhDUUIjSUSDAaNGJLMZJI799Fzb2KEFgriwgPffOfemXu+c8+Z
YymlcBcL4I5mXzuZTOat7/trhAZZrV6vVz3PKxJStVrtgrhMfEzYTiaT/J8A9ME6HQQhmUgkNm6d
geM4D0n1MwXavNUVvp6IU+KI63qgAAvdz+ZKO78a4EISBHwuwH0BRuz7es0Ns6af1xlEQnYVGJ0y
EZVyoBsjNeihuySIpdR7D8i3yLeJgYP0VaTdhYHwI+QvXAw+6ULuvIqRaBeyZxWM9XXjOF9GfKAH
6VwJL4Z7zZ7OkDEOa/83UyHbQb7oGkX9QitygjDX0FBtjseCmBgKYu9HAbs/iwhwLs3BmfHHmJ0M
010lFl5G0WACbxJ9qDOJpVcxWkuszA+Zw9p0QF0P69ORq8IdHk4KFXCulW9UdQZ+mykzcZPZ8/4g
tlMF2KbKpPp6+ik+7GWxONP/37a9e5/C6nIcHw9PjZC19e1KRTsZRKvqUkpT4TYr1dpXxteTo2ST
v3w/g91g/DLtqt6/+tv0W/+AyVKIZoEJktAawUvr3qfxjwADAPOBecF+sUNmAAAAAElFTkSuQmCC'! !

!MultiPointerHandMorph class methodsFor: 'misc' stamp: 'test 5/20/2009 14:39'!
scaledCursor: size colored: aColor

	| form canvas leftPoint rightPoint barOffset borderWidth bo |
	form := Form extent: size@size depth: 32.
	canvas := FormCanvas on: form.
	borderWidth := ((size / 20) max: 1) rounded.
	bo := borderWidth@borderWidth.
	
	rightPoint := Point r: size*0.75 degrees: 45.
	leftPoint := Point r: size*0.75 degrees: 90.
	barOffset := Point r: size*0.2 degrees: 62.5.
	
	canvas
		drawPolygon: {
			bo.
			bo + rightPoint.
			bo + (leftPoint * 0.33) + (rightPoint * 0.67).
			bo + (leftPoint * 0.33) + (rightPoint * 0.67) + barOffset.
			bo + (leftPoint * 0.67) + (rightPoint * 0.33) + barOffset.
			bo + (leftPoint * 0.67) + (rightPoint * 0.33).
			bo + leftPoint} 
		fillStyle: aColor
		borderWidth: borderWidth
		borderColor: Color black.
	
	^ (form asCursorForm) offset: bo negated; yourself! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/29/2009 17:49'!
buttons

	^ buttons! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/6/2009 22:21'!
buttons: newButtons

	buttons := newButtons.! !

!MultiPointerHandMorph methodsFor: 'updating' stamp: 'mig 5/19/2009 20:53'!
changed

	bounds := bounds origin extent: self currentCursor extent.
	super changed.! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 20:04'!
createMorphicButtonEvents: aState

	| result |
	result := OrderedCollection new.
	
	0 to: 2 do: [:each | | buttonBit |
		buttonBit := 1 bitShift: (2-each).
	
		(aState buttons anyMask: (1 bitShift: each*2)) ifTrue: [
			self buttons: (self buttons bitOr: buttonBit).
			result add: (self pvtMouseButtonEvent: #mouseDown with: buttonBit)].
	
		(aState buttons anyMask: (1 bitShift: (each*2)+1)) ifTrue: [
			self buttons: (self buttons bitAnd: (buttonBit bitInvert)).
			result add: (self pvtMouseButtonEvent: #mouseUp with: buttonBit)]].
	
	^ result! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/27/2009 10:54'!
createMorphicMoveEvent: aState

	| newPosition evt delta |
	delta := aState x @ aState y.
	delta isZero ifTrue: [^nil].
	
	newPosition := (self lastPosition + delta) adhereTo: Display fullBoundingBox.
	
	evt := MouseMoveEvent new
		setType: #mouseMove
		startPoint: self lastPosition
		endPoint: newPosition
		trail: nil
		buttons: self buttons
		hand: self
		stamp: Time millisecondClockValue.
	
	(self lastPosition = newPosition) ifFalse: [
		self resetTimeout].
	self lastPosition: newPosition.
	^ evt! !

!MultiPointerHandMorph methodsFor: 'cursor' stamp: 'mig 5/18/2009 14:18'!
currentCursor

	^ {self temporaryCursor. self normalCursor. NormalCursor} detect: [:cursor| cursor notNil]! !

!MultiPointerHandMorph methodsFor: 'drawing' stamp: 'mig 5/27/2009 10:40'!
drawOn: aCanvas 
	"Draw the hand itself (i.e., the cursor)."
	
	self hasMovedInLastTime ifFalse: [^ self].
	
	aCanvas 
		paintImage: self currentCursor
		at: bounds topLeft.

	self = MultiPointerManager instance keyboardHand ifTrue: [
		aCanvas
			paintImage: self class keyboardIcon
			at: (bounds bottomRight - self class keyboardIcon extent)].! !

!MultiPointerHandMorph methodsFor: 'event handling' stamp: 'SR 5/5/2009 23:03'!
enqueueEvent: aState

	aState isMorphicEvent
		ifTrue: [self queue nextPut: aState]
		ifFalse: [
			self queue nextPut: (self createMorphicMoveEvent: aState).
			aState buttons isZero ifFalse: [
				self queue nextPutAll: (self createMorphicButtonEvents: aState)]].! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 16:15'!
handColor

	^ handColor! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 20:06'!
handColor: aColor

	handColor := aColor.
	self normalCursor: (self class cursorColored: aColor).! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:28'!
handId

	^ self pointer pointerID asString
! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/27/2009 10:56'!
hasMovedInLastTime

	^ self movedTimestamp notNil 
		and: [Time millisecondClockValue - self movedTimestamp < self class displayTimeout]! !

!MultiPointerHandMorph methodsFor: 'initialization' stamp: 'mig 5/27/2009 10:26'!
initialize

	super initialize.
	self
		queue: SharedQueue new;
		lastPosition: 0@0;
		buttons: 0.
! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/24/2009 13:38'!
lastPosition

	^ lastPosition! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 20:06'!
lastPosition: anObject

	lastPosition := anObject.! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/27/2009 10:25'!
movedTimestamp

	^ movedTimestamp! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/27/2009 10:25'!
movedTimestamp: anObject

	movedTimestamp := anObject.! !

!MultiPointerHandMorph methodsFor: 'drawing' stamp: 'SR 4/24/2009 13:39'!
needsToBeDrawn

	^ true! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/6/2009 10:41'!
normalCursor

	^ normalCursor! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 20:53'!
normalCursor: anObject

	normalCursor := anObject.
	self changed.! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:22'!
pointer

	^ pointer! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:22'!
pointer: anObject

	pointer := anObject.! !

!MultiPointerHandMorph methodsFor: 'event handling' stamp: 'mig 5/7/2009 13:54'!
processEvents
	
	ActiveEvent ifNotNil: [
		"Meaning that we were invoked from within an event response. Make sure z-order is up to date"
		self mouseOverHandler processMouseOver: lastMouseEvent].
	
	self queue isEmpty ifTrue: [
		"No pending events. Make sure z-order is up to date"
		self mouseOverHandler processMouseOver: lastMouseEvent.
		^ self].

	[self queue isEmpty] whileFalse: [
		self handleEvent: self queue next].! !

!MultiPointerHandMorph methodsFor: 'private' stamp: 'mig 5/19/2009 20:04'!
pvtMouseButtonEvent: aSymbol with: anInteger

	^ MouseButtonEvent new 
		setType: aSymbol
		position: self lastPosition
		which: anInteger
		buttons: self buttons
		hand: self
		stamp: Time millisecondClockValue! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/24/2009 13:38'!
queue

	^ queue! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/6/2009 22:21'!
queue: anObject

	queue := anObject.! !

!MultiPointerHandMorph methodsFor: 'event handling' stamp: 'mig 5/19/2009 10:27'!
readEvent

	self enqueueEvent: self pointer state.! !

!MultiPointerHandMorph methodsFor: 'event handling' stamp: 'mig 5/27/2009 10:55'!
resetTimeout

	self movedTimestamp: Time millisecondClockValue.! !

!Cursor class methodsFor: '*xpf-multipoint-core-override' stamp: 'mig 5/18/2009 15:03'!
extent: extentPoint fromArray: anArray offset: offsetPoint 
	"Answer a new instance of me with width and height specified by
	extentPoint, offset by offsetPoint, and bits from anArray.
	The width must be smaller or equal to 32 to fit into an 32-bit bitmap"
	
	(extentPoint x between: 1 and: 32) ifFalse: [
		^ self error: 'cursors must be max. 32 pixels wide'].

	^ super
		extent: extentPoint
		fromArray: (anArray collect: [:bits | bits bitShift: 32 - extentPoint x])
		offset: offsetPoint.
! !

!SharedQueue methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 5/4/2009 10:34'!
nextPut: value 
	"Send value through the receiver. If a Process has been suspended 
	waiting to receive a value through the receiver, allow it to proceed."

	value ifNil: [^ self].
	accessProtect
		critical: [writePosition > contentsArray size
						ifTrue: [self makeRoomAtEnd].
				 contentsArray at: writePosition put: value.
				 writePosition := writePosition + 1].
	readSynch signal.
	^value! !

!SharedQueue methodsFor: '*xpf-multipoint-core' stamp: 'SR 4/29/2009 17:24'!
nextPutAll: aCollection

	aCollection do: [:each | self nextPut: each].! !

!MenuItemMorph methodsFor: '*xpf-multipoint-core-override' stamp: 'bs 5/20/2009 20:12'!
mouseDown: evt
	"Handle a mouse down event. Menu items get activated when the mouse is over them."

	evt shiftPressed ifTrue: [^ super mouseDown: evt].  "enable label editing" 

	"Quick hack to bring menu to the front if it is obscured. See 
		http://bugs.squeak.org/view.php?id=1780" "fixed by gvc to work with embedded menus"
	"self owner owner = self world ifTrue: [
		Transcript cr; show: self owner asString. 
		self world morphsInFrontOf: self owner overlapping: self owner bounds
			do: [:ignored | ^ self owner comeToFront]]."

	
	(self isInDockingBar
			and:[isSelected]
			"and:[owner selectedItem == self]")
		ifTrue:[
			evt hand newMouseFocus: nil.
			owner selectItem: nil event: evt. ]
		ifFalse:[
			evt hand newMouseFocus: owner. "Redirect to menu for valid transitions"
			owner selectItem: self event: evt. ]
! !

!Morph methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 5/6/2009 14:25'!
delete
	"Remove the receiver as a submorph of its owner and make its 
	new owner be nil."

	| aWorld |
	self removeHalo.
	aWorld := self world ifNil: [World].
	"Terminate genie recognition focus"
	"I encountered a case where the hand was nil, so I put in a little 
	protection - raa "
	" This happens when we are in an MVC project and open
	  a morphic window. - BG "
	aWorld ifNotNil: [
		self disableSubmorphFocusForHand: self activeHand.
		self activeHand ifNotNil: [
			self activeHand
				releaseKeyboardFocus: self;
				 releaseMouseFocus: self]].
	owner ifNotNil:[ self privateDelete.
		self player ifNotNilDo: [ :player |
			"Player must be notified"
			player noteDeletionOf: self fromWorld: aWorld]].! !
MultiPointerManager initialize!
